// Copyright 1994-2002 by Jon Dart. All Rights Reserved.
#pragma optimize("g",off)

#include "stdafx.h"
#include "display.h"
#include "square.h"
#include "resource.h"
#include "arasan.h"

#define STATUS_AREA_WIDTH       122
#define TEXT_OFFSET             30 /* from right edge */
#define COORD_OFFSET            5
#define TIME_Y                  40
#define MOVE_Y                  198
#define STATUS_Y                116
#define PLAYERS_Y               132
#define RESULT_Y                215
#define ECO_Y                   270

static const LOGFONT textLF = { 8,0,0,0,700,0,0,0,ANSI_CHARSET,
  OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
  DEFAULT_PITCH | FF_SWISS,
  "MS Sans Serif"};

static const LOGFONT coordLF = { 6,0,0,0,700,0,0,0,ANSI_CHARSET,
  OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
  DEFAULT_PITCH | FF_SWISS,
  "MS Sans Serif"};

// Piece map for "Chess Alpha" font:
static const char *whiteSquareMapAlpha = " ojntwl  phbrqk ";
static const char *blackSquareMapAlpha = "+OJNTWL++PHBRQK+";

// Piece map for "Marroquin", "Maya" and similar fonts:
static const char *whiteSquareMapMaya = " omvtwl  pnbrqk ";
static const char *blackSquareMapMaya = "+OMVTWL++PNBRQK+";

// Piece map for "Berlin" font:
static const char *whiteSquareMapBerlin = " ojntwl  phbrqk ";
static const char *blackSquareMapBerlin = "+OJNTWL++PHBRQK+";

static const char **blackSquareMap;
static const char **whiteSquareMap;

static CFont *textFont = NULL;
static CFont *coordFont = NULL;

static DWORD square_width,square_height;
DWORD Display::spacing;

static void clearRect(CDC *pDC, CRect &rect)
{
     CRgn rgn;
     rgn.CreateRectRgnIndirect(rect);
     CBrush brush;
     brush.CreateSolidBrush(GetSysColor(COLOR_WINDOW));
     pDC->FillRgn(&rgn,&brush);
}

Display::Display( CWnd *pWin, const CRect &initialSize )
: turned(FALSE)
{
     CDC *pDC = pWin->GetDC();
     myWin = pWin;
     int fontSize;
     switch (guiOptions->getBoardSize()) {
     case GuiOptions::Small: fontSize = 240; break;
     case GuiOptions::Medium: fontSize = 360; break;
     case GuiOptions::Large: fontSize = 480; break;
     }
     if (!pieceFont.CreatePointFont(fontSize,guiOptions->getPieceFontName(),pDC)) {
         AfxMessageBox("failed to load chess font");
         pWin->ReleaseDC(pDC);
         return;
     }
     
    TEXTMETRIC my_tm;

     textFont = new CFont();

     textFont->CreateFontIndirect(&textLF);

     coordFont = new CFont();
     coordFont->CreateFontIndirect(&coordLF);

     pDC->SelectObject(textFont);
 
     pDC->GetTextMetrics(&my_tm);
     mono = (pDC->GetDeviceCaps(NUMCOLORS) <= 2);
	 if (mono)
	 {
        // 16- and 32-bit color modes do not seem
		// to report NUMCOLORS correctly.
		mono = pDC->GetDeviceCaps(BITSPIXEL) == 1 &&
			pDC->GetDeviceCaps(PLANES) == 1;

	 }
     char_width = my_tm.tmAveCharWidth;
     vertical_border_width = char_width + 7;
     spacing = my_tm.tmHeight + my_tm.tmExternalLeading;

     // now select piece font to get data from that
     updatePieceFont(pDC);
     pWin->ReleaseDC(pDC);
     setSize(initialSize);
}

void Display::setPieceFont(CDC *pDC, LPCSTR fontName, GuiOptions::BoardSize boardSize)
{
     CFont newFont;
     int fontSize;
     switch(boardSize) {
     case GuiOptions::Small: fontSize = 240; break;
     case GuiOptions::Medium: fontSize = 360; break;
     case GuiOptions::Large: fontSize = 480; break;
     }
     if (!newFont.CreatePointFont(fontSize,fontName)) {
          ::MessageBox(NULL,"Cannot create font!","Error",MB_ICONEXCLAMATION);
          return;
     }
     LOGFONT lf;
     newFont.GetLogFont(&lf);
     setPieceFont(pDC,&lf);
}

void Display::resize(CDC *pDC, GuiOptions::BoardSize size) {
    LOGFONT lf;
    pieceFont.GetLogFont(&lf);
    setPieceFont(pDC,lf.lfFaceName,size);
}

void Display::updatePieceFont(CDC *pDC) {
    TEXTMETRIC piece_tm;
    pDC->SelectObject(pieceFont);
    pDC->GetTextMetrics(&piece_tm);
     
    square_width = piece_tm.tmMaxCharWidth;
    square_height = piece_tm.tmHeight;
    horiz_border_height = square_height + 4;
    board_right_edge = piece_tm.tmMaxCharWidth*8 + 50;
    textX = 8*square_width+TEXT_OFFSET;
    coordX = 8*square_width+COORD_OFFSET;
    LOGFONT lf;
    pieceFont.GetLogFont(&lf);
    if (strcmp(lf.lfFaceName,"Chess Alpha")==0) {
       blackSquareMap = (const char**)&blackSquareMapAlpha;
       whiteSquareMap = (const char**)&whiteSquareMapAlpha;
    }
    else if (strcmp(lf.lfFaceName,"Chess Berlin")==0) {
       blackSquareMap = (const char**)&blackSquareMapBerlin;
       whiteSquareMap = (const char**)&whiteSquareMapBerlin;
    }
    else {
       blackSquareMap = (const char**)&blackSquareMapMaya;
       whiteSquareMap = (const char**)&whiteSquareMapMaya;
    }
}

Display::~Display() {
   delete textFont;
   delete coordFont;
}

void Display::drawBoard( CDC *pDC, const Board &board, const CRect &drawArea, BOOL turned) {

     pDC->SelectObject(pieceFont);
     int vert = 0;
     for (int i = 0; i < 8; i++)
     {
        char row[8];
        for (int j = 0; j < 8; j++)
        {
           Square sq;
           if (turned)
               sq = Square(8-j,8-i,Black);
           else
               sq = Square(j+1,i+1,Black);
           if (SquareColor(sq) == White)
               row[j] = (*whiteSquareMap)[board[sq]];
           else
               row[j] = (*blackSquareMap)[board[sq]];
        }
        CRect loc(0,vert,8*square_width,vert+square_height);
        pDC->DrawText(row,8,&loc,DT_LEFT); 
        vert += square_height;
     }
     showSide(pDC,board.Side());
}

void Display::drawPiece(CDC *pDC, Square sq, Piece p)
{
     int f = File(sq);
     int r = Rank(sq,Black);
     if (turned) {
       r = 8-r;
       f = 8-r;
     }
     char text;
     
     if (SquareColor(sq) == White)
        text = (*whiteSquareMap)[p];
     else
        text = (*blackSquareMap)[p];
     int vert = square_height*(r-1);
     CRect loc(0,vert,(f-1)*square_width,vert+square_height);
     pDC->DrawText(&text,1,&loc,DT_LEFT); 
}


void Display::setSize(const CRect &size) {
   width = size.Width();
}

void Display::showSide(CDC *pDC, ColorType side) {

    CFont *oldFont = (CFont*)pDC->SelectObject(textFont);
    CRect loc(textX,12,width-5,52);
    if (side == White) 
      pDC->DrawText("WHITE to Move",13,&loc,DT_LEFT);
    else
      pDC->DrawText("BLACK to Move",13,&loc,DT_LEFT);
    pDC->SelectObject(oldFont);
}

void Display::showTime(HWND hWin, time_t time, ColorType side)
{
    if (!textFont) // not initialized yet
        return;

    // internal time is in 1/100 second increments
    unsigned hours = (unsigned)(time/360000U);
    unsigned minutes = (unsigned) ((time - (hours*360000U))/6000U);
    unsigned seconds = (unsigned)(time - (hours*360000U) - (minutes*6000U))/100;
    // convert time to ASCII:
    char time_str[50];
    wsprintf(time_str,"%02d:%02d:%02d",
       (int)hours,(int)minutes,(int)seconds);

     //  Note: we use straight Windows code - not MFC classes - because
     // this method may be called from a different thread, and MFC
     // CWnd is not designed to be shared across threads.
     HDC hdc = GetDC(hWin);
     ASSERT(hdc);
     SelectObject(hdc,textFont->m_hObject);

     RECT rcText, client_rect, text_rect;
     GetClientRect( hWin, &client_rect );
     int width = client_rect.right - client_rect.left;
     int vbase = TIME_Y;
 	 SetBkColor(hdc,GetSysColor(COLOR_WINDOW));

     int x = board_right_edge + 5;
     if (side == White)
     {
        SetRect(&text_rect,x,vbase,width-5,vbase+spacing);
        DrawText(hdc,"WHITE:",6,&text_rect,DT_LEFT);
        SetRect(&rcText,x,vbase+spacing,width-5,vbase+2*spacing);
     }
     else
     {
        SetRect(&text_rect,x,vbase+2*spacing,width-5,
                vbase+3*spacing);
        DrawText(hdc,"BLACK:",6,&text_rect,DT_LEFT);
        SetRect(&rcText,x,vbase+3*spacing,width-5,vbase+4*spacing);
     }
     DrawText(hdc,time_str,lstrlen(time_str),&rcText,
                DT_LEFT | DT_WORDBREAK);
	 
     BOOL ret = ReleaseDC(hWin,hdc);
     ASSERT(ret);
}

void Display::showECO(CDC *pDC, const char *eco, const char *openingName)
{
     if (!textFont) // not initialized yet
         return;
     pDC->SelectObject(textFont);
     CRect rcText(textX,ECO_Y,width-5,ECO_Y+spacing);
     clearRect(pDC,rcText);
	 COLORREF oldColor = pDC->SetBkColor(GetSysColor(COLOR_WINDOW));

     pDC->DrawText(eco,strlen(eco),&rcText,DT_LEFT);
     rcText.SetRect(textX,ECO_Y+spacing,width-5,ECO_Y+4*spacing);
     clearRect(pDC,rcText);

     pDC->DrawText(openingName,strlen(openingName),
                   &rcText, DT_LEFT | DT_WORDBREAK);
	 pDC->SetBkColor(oldColor);

}


void Display::showMove(class CDC *pDC,const string &image,int moveCount,ColorType side) {

     pDC->SelectObject(textFont);
     CRect rcText(textX,(int)MOVE_Y,textX+width-5,
      (int)MOVE_Y+spacing);
     char text[80]; 
     // erase any previous text:
     clearRect(pDC,rcText);
     if (image.length()==0) {
       return;
     }
     int move_num = (moveCount-1)/2;
     if (side == White)
       wsprintf(text,"%d  %s",move_num+1,image.c_str());
     else
       wsprintf(text,"%d ... %s",move_num+1,image.c_str());
	 COLORREF oldColor = pDC->SetBkColor(GetSysColor(COLOR_WINDOW));
     pDC->DrawText(text,strlen(text), &rcText, DT_LEFT | DT_WORDBREAK);
	 pDC->SetBkColor(oldColor);
}

void Display::showResult(CDC *pDC, const string &result) {
     pDC->SelectObject(textFont);
     CRect rcText(textX,(int)RESULT_Y,width-5,
      (int)RESULT_Y+spacing);
     // erase any previous text:
     clearRect(pDC,rcText);
     if (result.length()==0 || (result.compare("*")==0)) {
       return;
     }
	 COLORREF oldColor = pDC->SetBkColor(GetSysColor(COLOR_WINDOW));
     pDC->DrawText(result.c_str(),result.length(), &rcText, DT_LEFT | DT_WORDBREAK);
	 pDC->SetBkColor(oldColor);     
}

void Display::showCoordinates(CDC *pDC) {
     CFont *oldFont = pDC->SelectObject(coordFont);
     TEXTMETRIC tm;
     pDC->GetTextMetrics(&tm);
     int fontHeight = tm.tmHeight + tm.tmExternalLeading;
     CString letters;
     letters.LoadString(IDS_LETTER_COORDS);
     CString numbers;
     numbers.LoadString(IDS_NUMBER_COORDS);
     int i;
     for (i = 0; i < 8; i++) {
        CRect loc(i*square_width,square_height*8+6,(i+1)*square_width,square_height*8+6+fontHeight);
        int midY = i*square_height + square_height/2;
        CRect loc2(coordX,midY-fontHeight/2,coordX+18,midY+fontHeight/2);
        if (turned) {
           pDC->DrawText((LPCSTR)letters.Mid(7-i,1),1,&loc,DT_CENTER); 
           pDC->DrawText((LPCSTR)numbers.Mid(i,1),1,&loc2,DT_CENTER | DT_VCENTER ); 
        }
        else {
           pDC->DrawText((LPCSTR)letters.Mid(i,1),1,&loc,DT_CENTER); 
           pDC->DrawText((LPCSTR)numbers.Mid(7-i,1),1,&loc2,DT_CENTER | DT_VCENTER ); 
        }
     }
     pDC->SelectObject(oldFont);
}

Square Display::mouseLocation(const CPoint &p) const
{
     int x,y;
     x = (p.x/square_width)+1;
     if (turned)
        x = 9-x;
     y = (p.y/square_height)+1;
     if (x > 8 || y > 8)
        return InvalidSquare;
     else if (turned)
        return Square(x,y,White);
	 else
		return Square(x,y,Black);    
}

void Display::showHeader(CDC *pDC, const string &header)
{
    if (!textFont) // not initialized yet
        return;
    pDC->SelectObject(textFont);
    CRect rcText(textX,PLAYERS_Y,width-5,PLAYERS_Y+4*spacing);
    clearRect(pDC,rcText);
    COLORREF oldColor = pDC->SetBkColor(GetSysColor(COLOR_WINDOW));

    pDC->DrawText(header.c_str(),header.length(),
                  &rcText, DT_LEFT | DT_WORDBREAK | DT_NOPREFIX);
	pDC->SetBkColor(oldColor);
}

void Display::calcWindowSize(GuiOptions::BoardSize boardSize, int &w, int &h)
{
   switch (boardSize) {
      case GuiOptions::Small: w = 500; h = 475; break;
      case GuiOptions::Medium: w = 700; h = 650; break;
      case GuiOptions::Large: w = 900; h = 875; break;
   }

}


