// guiDoc.cpp : implementation of the ArasanGuiDoc class
// Copyright 2001, 2002 by Jon Dart. All Rights Reserved.

#include "stdafx.h"
#include "arasan.h"
#include "EngineCo.h"

#include "guiDoc.h"
#include "chessio.h"
#include "selectga.h"
#include "notation.h"
#include "boardio.h"
#include "legal.h"
#include "guiview.h"
#include "mainfrm.h"
#include <fstream>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// ArasanGuiDoc

IMPLEMENT_DYNCREATE(ArasanGuiDoc, CDocument)

BEGIN_MESSAGE_MAP(ArasanGuiDoc, CDocument)
	//{{AFX_MSG_MAP(ArasanGuiDoc)
	ON_COMMAND(ID_FILE_NEW, OnFileNew)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
    ON_COMMAND_EX_RANGE(ID_FILE_MRU_FILE1, ID_FILE_MRU_FILE16, OnRecentFile)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////

// ArasanGuiDoc construction/destruction


ArasanGuiDoc::ArasanGuiDoc()
: currentMove(0), pgn_file(NULL),
  lastPing(0), lastPong(0), computerSide(Black)
{
    ecoCoder = new ECO();
}

ArasanGuiDoc::~ArasanGuiDoc()
{
    delete ecoCoder;
}

BOOL ArasanGuiDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

    currentMove = 0;
    sendNew();
    reset();

	return TRUE;
}

void ArasanGuiDoc::getECO(string &eco,string &openingMove) {
    ecoCoder->classify(moveList,eco,openingMove);
}

// Update the board after a move is made
void ArasanGuiDoc::updateBoard(const Move m) {
    Board_State previous_state(currentBoard);
    char image[20];
    if (!IsNull(m)) {
        POSITION pos = GetFirstViewPosition();
        ArasanGuiView* pFirstView = (ArasanGuiView*)GetNextView( pos );
        pFirstView->update_board(m);
        Notation::Image(currentBoard,m,image);
        currentBoard.DoMove(m);
        moveList.truncate(currentMove);
        moveList.add_move(currentBoard,previous_state,m,image);
        currentMove = numMoves();
        pFirstView->Invalidate();
    }
    // No need to mark us as modified since we already updated the view
    SetModifiedFlag(FALSE);
}

void ArasanGuiDoc::goToMove(const int index) {
    ASSERT(index >= 0 && index <= moveList.length());
    if (index == currentMove)
        return;
    setForce(1);
    // Note: we can't use "setboard" here because the engine
    // needs to have the previous game context. So we force
    // the engine to step back/forward through the moves:
    if (index > currentMove) {
        do {
            const MoveRecord &rec = moveList[currentMove];
            char image[50],cmd[50];
            Notation::Image(currentBoard,rec.move(),image);
            currentBoard.DoMove(rec.move());
            sprintf(cmd,"usermove %s",image);
            writeToEngine(cmd);
            ++currentMove;
        } while (index != currentMove);
    }
    else {
        while (index != currentMove) {
            const MoveRecord &rec = moveList[currentMove-1];
            currentBoard.UndoMove(rec.move(),rec.state());
            writeToEngine("undo");
            --currentMove;
        }
    }
    ArasanMainFrame* pMainFrm = (ArasanMainFrame*)AfxGetMainWnd();
    if (pMainFrm) pMainFrm->clearStatusBar();
    SetModifiedFlag();
    UpdateAllViews(NULL);
}

void ArasanGuiDoc::goForward(const Move m) {
    // This is called when browsing forward through a game.
    // N.b. can't call updateBoard as it will truncate the move list.
    Board_State previous_state(currentBoard);
    char image[20];
    if (!IsNull(m)) {
        POSITION pos = GetFirstViewPosition();
        ArasanGuiView* pFirstView = (ArasanGuiView*)GetNextView( pos );
        pFirstView->update_board(m);
        Notation::Image(currentBoard,m,image);
        char buf[128];

        // make sure we're in "force" mode
        setForce(1);
        // send the move to the engine
        sprintf(buf,"usermove %s\n",image);
        writeToEngine(buf);

        currentBoard.DoMove(m);
        currentMove++;
        pFirstView->Invalidate();
    }
    ArasanMainFrame* pMainFrm = (ArasanMainFrame*)AfxGetMainWnd();
    if (pMainFrm) pMainFrm->clearStatusBar();
    // No need to mark us as modified since we already updated the view
    SetModifiedFlag(FALSE);
}

void ArasanGuiDoc::undoMove() {
    // get last record
    if (currentMove==0) return;
    MoveRecord rec = moveList[currentMove-1];
    setForce(1);
    writeToEngine("undo\n");
    currentBoard.UndoMove(rec.move(),rec.state());
    currentMove--;
    ArasanMainFrame* pMainFrm = (ArasanMainFrame*)AfxGetMainWnd();
    if (pMainFrm) pMainFrm->clearStatusBar();
    SetModifiedFlag();
    UpdateAllViews(NULL);
}


/////////////////////////////////////////////////////////////////////////////
// ArasanGuiDoc serialization

void ArasanGuiDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}

void ArasanGuiDoc::loadBoard(const Board &newBoard) {
    currentBoard = newBoard;
    while(numMoves())
      moveList.remove_move();
    currentMove = 0;
    SetModifiedFlag(FALSE);
    UpdateAllViews(NULL);
}

void ArasanGuiDoc::sendNew() {
    if (engineReady()) {
       // We may get incoming stuff from the engine, including a move,
       // between the time we issue the "new" command here and the
       // time it's executed. So use the "ping" mechanism to enable
       // us to reject extra commands.
       ++lastPing;
       writeToEngine("new\n");
       // "new" will reset force mode in the engine
       extern int force;
       force = 0;
       char msg[30];
       sprintf(msg,"ping %d\n",lastPing);
       writeToEngine(msg);
    }
}

void ArasanGuiDoc::reset()
{
    Board cleanBoard;
    loadBoard(cleanBoard);
    pgnContents.removeAll();
    currentMove = 0;
    computerSide = Black;
    players = "";
    result = "*"; // means "incomplete"
    POSITION pos = GetFirstViewPosition();
    // Note: there is only one view (SDI)
    ArasanGuiView* pFirstView = (ArasanGuiView*)GetNextView( pos );
    SetModifiedFlag(FALSE);
    ArasanMainFrame* pMainFrm = (ArasanMainFrame*)AfxGetMainWnd();
    if (pMainFrm) pMainFrm->clearStatusBar();
    pFirstView->reset();
}

BOOL ArasanGuiDoc::load_pgn(Board &board,ifstream &game_file, DWORD offset,
ArasanVector<StringPair> &hdrs)
{
    // Make sure engine removes old game moves
    // and is at start of game.
    sendNew();
    setForce(1);
    board.Reset();
    ColorType side = White;
    game_file.clear();
    game_file.seekg(offset); 

    // Collect the header:
    long first;
    ChessIO::collect_headers(game_file, hdrs, first);

    // See if the file contained a "FEN" tag.
    string FENString;
    if (ChessIO::get_header(hdrs,"FEN",FENString))
    {
       const char *buf = FENString.c_str();

       istrstream FENstream(buf,256);
       if (!ChessIO::load_fen(FENstream,board))
       {
          CString msg;
          msg.LoadString(IDS_INVALID_FEN);
          ::MessageBox(NULL,(LPCSTR)msg,"",MB_OK);
          return FALSE;
       }
    }

    // read the moves
    currentMove = 0;
    moveList.removeAll();
    
    Board tmp_board(board);
    for (;;)
    {
        static char buf[256];
        static char num[20];
        ChessIO::Token tok = ChessIO::get_next_token(game_file,buf,256);
        if (tok == ChessIO::Eof)
             break;
        if (tok == ChessIO::Number)
        {
            strncpy(num,buf,20);
        }
        else if (tok == ChessIO::GameMove)
        {
            // parse the move
            Move m = Notation::Value(tmp_board,side,buf);
            if (IsNull(m) ||
                !legal_move(tmp_board,StartSquare(m),
                            DestSquare(m)))
            {
                CString msg;
                AfxFormatString1(msg,IDS_ILLEGAL_MOVE,buf);
                ::MessageBox(NULL,(LPCSTR)msg,"",MB_OK);
                break;
            }
            else
            {
                Board_State previous_state(tmp_board);
                char image[20];
                // Don't use the current move string as the input
                // parser is forgiving and will accept incorrect
                // SAN. Convert it here to the correct form:
                Notation::Image(tmp_board,m,image);
                moveList.add_move(tmp_board,previous_state,m,image);
                tmp_board.DoMove(m);
            }
            side = OppositeColor(side);
         }
        else if (tok == ChessIO::Unknown)
         {
             CString msg;
             AfxFormatString1(msg,IDS_UNREC_TEXT,buf);
             ::MessageBox(NULL,(LPCSTR)msg,"",MB_OK);
         }
         else if (tok == ChessIO::Comment)
         {
             // ignored for now
         }
         else if (tok == ChessIO::Result)
         {
            //if (*buf != '*')
            //   log.setResult(buf);
            break;
         }            
    } 
    // position us at the start of the game
    currentMove = 0;
    UpdateAllViews(NULL);
    return TRUE;
}

const char *ArasanGuiDoc::lastMoveImage() const {
    if (currentMove == 0) return "";
    return moveList[currentMove-1].image();
}

void ArasanGuiDoc::setResult(const string &res) {
    result = res;
    char buf[200];
    strcpy(buf,"result ");
    strcat(buf,res.c_str());
    writeToEngine(buf);
    POSITION pos = GetFirstViewPosition();
    ArasanGuiView* pFirstView = (ArasanGuiView*)GetNextView( pos );
    pFirstView->showResult(result);
}

BOOL ArasanGuiDoc::open_file(LPCSTR fileName) {
   // framework may call this with fileName == NULL, in
   // response to OnFileNew
   if (!fileName)
   {
      // TBD OnNewGame();
      return FALSE;
   }
   BOOL status = FALSE;
   // get the extension:
   CString ext;
   CString name(fileName);
   int i = name.GetLength()-1;
   while (i>=0 && name[i] != '.') --i;
   if (name[i] == '.')
      ext = name.Right(name.GetLength()-i-1);
//   free_pgn();
   pgn_file = new ifstream();

   pgn_file->open(fileName,ios::in | ios::binary);
   if (!pgn_file->is_open() || !pgn_file->good())
   {
      CString msg;
      AfxFormatString1(msg,IDS_CANT_OPEN,fileName);
      AfxMessageBox(msg);
      return FALSE;
   }

   ext.MakeLower();
   BOOL multipleGameFile = FALSE;
   if (ext == "pgn")
   {
      SetPathName(fileName);

      currentBoard.Reset();
      pgnContents.removeAll();     

      // might take awhile, put up a wait cursor. A
      // progress bar would be nicer ...
      BeginWaitCursor();
      // scan just the PGN headers, so we can tell the
      // user what's in the file:
      ChessIO::scan_pgn(*pgn_file,pgnContents);
      EndWaitCursor();

      DWORD pgnOffset = 0L;
      if (pgnContents.length() > 1)
      {
         // let the user choose a game to view
        POSITION pos = GetFirstViewPosition();
        CView* pFirstView = GetNextView( pos );

         SelectGame dlg(pgnContents,pFirstView);
         if (dlg.DoModal() == IDCANCEL)
         {
            return FALSE;  
         }
         selectGame(dlg.get_selection());
         return TRUE;
      }
      headers.removeAll();
      status = load_pgn(currentBoard,*pgn_file,pgnOffset,headers);
	  ChessIO::get_game_description(headers,players);
      SetModifiedFlag(FALSE);
      UpdateAllViews(NULL);
   }
   else if (ext == "fen")
   {
      status = ChessIO::load_fen(*pgn_file,currentBoard);
      if (!status)
      {
         CString msg;
         AfxFormatString2(msg,IDS_INVALID_BOARD,fileName,ext);
         AfxMessageBox(msg);   
         currentBoard.Reset();

      }
      else
      {
         char buf[200];
         // Sync engine with new board position
         ostrstream s(buf,256);
         s << "setboard " << currentBoard;
         writeToEngine(s.str());
         SetPathName(fileName);
         CString msg;
         AfxFormatString1(msg,IDS_BOARD_LOADED,fileName);
         SetModifiedFlag();
         UpdateAllViews(NULL);
      }
   }
   else
   {
      CString msg;
      msg.LoadString(IDS_UNKNOWN_FILE_TYPE);
      AfxMessageBox((LPCSTR)msg);
      return FALSE;
   }
   // If we read a pgn file consisting of >1 game, keep the file
   // open:
   if (!multipleGameFile)
   {
      if (pgn_file) pgn_file->close();
      delete pgn_file;
      pgn_file = NULL;
   }
   POSITION pos = GetFirstViewPosition();
   ArasanGuiView* theView = (ArasanGuiView*)GetNextView( pos );
   theView->fileLoaded();
   // We have updated the view
   SetModifiedFlag(FALSE);
   return TRUE;                       
}

/////////////////////////////////////////////////////////////////////////////
// ArasanGuiDoc diagnostics

#ifdef _DEBUG
void ArasanGuiDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void ArasanGuiDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// ArasanGuiDoc commands


void ArasanGuiDoc::OnCloseDocument() 
{
    POSITION pos = GetFirstViewPosition();
    ArasanGuiView* pView = (ArasanGuiView*)GetNextView( pos );
	pView->saveOptions();

	CDocument::OnCloseDocument();
}

void ArasanGuiDoc::OnFileNew() 
{
	OnNewDocument();
}

void ArasanGuiDoc::nextGame() {

   selectGame(++pgnSelection);
}

void ArasanGuiDoc::previousGame() {
   ASSERT(pgnSelection > 0);
   selectGame(--pgnSelection);
}

void ArasanGuiDoc::selectGame(DWORD index) {
   pgnSelection = index;
   const string &header = pgnContents[pgnSelection];
   string::size_type indx = header.find('\t');
   DWORD offset;
   string num = header.substr(indx+1,header.length()-indx-1);
   sscanf(num.c_str(),"%ld",&offset);
   ArasanVector<StringPair> hdrs;

   sendNew();
   Board cleanBoard;
   loadBoard(cleanBoard);

   load_pgn(currentBoard,*pgn_file,offset,hdrs);
   ChessIO::get_game_description(hdrs,players);

   currentMove = 0;
   computerSide = Black;
   POSITION pos = GetFirstViewPosition();
   // Note: there is only one view (SDI)
   ArasanGuiView* pFirstView = (ArasanGuiView*)GetNextView( pos );
   SetModifiedFlag(FALSE);
   ArasanMainFrame* pMainFrm = (ArasanMainFrame*)AfxGetMainWnd();
   if (pMainFrm) pMainFrm->clearStatusBar();
   pFirstView->reset(ArasanGuiView::NewPgn);
}

void ArasanGuiDoc::storeGame(ostream &ofs) {
   if (headers.length() > 0)
   {
      int ecoFound = 0;
      int n = headers.length();
      for (int i=0;i<n;i++)
      {
         const StringPair &p = headers[i];
         if (p.tag == "ECO")
         {
            ++ecoFound;
         }
      }
      if (!ecoFound)
         headers.append(StringPair("ECO",eco.c_str()));
   }
   else
   {
      headers.append(StringPair("ECO",eco.c_str()));
   }

   ChessIO::store_pgn(ofs,moveList,computerSide,result,headers);
}

void ArasanGuiDoc::OnFileSave() 
{
   CString filter;
   filter.LoadString(IDS_FILE_OPEN_FILTER);
   CFileDialog dlg(FALSE,
                   "pgn",  // default file extension
                   NULL,    // no default filename
                   OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
                   (LPCSTR)filter,
                   AfxGetMainWnd());
   if (dlg.DoModal() == IDOK)
   {
      // check the file extension
      CString str = dlg.GetFileExt();
      CString fileName = dlg.GetPathName();
      // Don't allow trailing spaces. Under NT these are part of the file name.
      // It is too easy to get these in due to typos.
      const int last = fileName.GetLength()-1;
      int j = 0;
      while (j<=last)
      {
         if (fileName[j] != ' ')
            break;
         ++j;
      }
      int i = last;
      while (i)
      {
         if (fileName[i] != ' ')
            break;
         i--;
      }
      fileName = fileName.Mid(j,i+1-j);
      ofstream ofs(fileName,ios::out);
      if (!ofs.good())
      {
         CString msg;
         AfxFormatString1(msg,IDS_CANT_OPEN,fileName);
         AfxMessageBox(msg);
         return;
      }
      str.MakeLower();
      str = str.Left(3);
      if (str == "pgn")
      {
         int ecoFound = 0;
         const int n = headers.length();

         for (int i=0;i<n;i++)
         {
            const StringPair &p = headers[i];
            if (p.tag == "ECO")
            {
                ++ecoFound;
                break;
			}
		 }
         // TBD if (!ecoFound)
         // TBD   headers.append(StringPair("ECO",eco));
         ChessIO::store_pgn(ofs,moveList,computerSide,result,headers);
      }
      else if (str == "fen" || str == "epd")
      {
         ChessIO::store_fen(ofs,currentBoard);
      }
      else
      {
         CString msg;
         AfxFormatString1(msg,IDS_UNKNOWN_FILE_TYPE,(LPCSTR)str);
         AfxMessageBox((LPCSTR)msg);         
      }
      ofs.close();
   }

}

BOOL ArasanGuiDoc::OnRecentFile(UINT nID)
{ 
   //ArasanGuiApp* pApp = (ArasanGuiApp*)AfxGetApp();
   //if (open_file(pApp->getRecentFile(nID - ID_FILE_MRU_FILE1))) {
   //  setForce(1);
   //}
   ArasanGuiApp* pApp = (ArasanGuiApp*)AfxGetApp();
   if (OnOpenDocument(pApp->getRecentFile(nID - ID_FILE_MRU_FILE1)))
       setForce(1);
   else
       return FALSE;
   return TRUE; // We processed this message
}


BOOL ArasanGuiDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
    // We don't use the default MFC method
    return open_file(lpszPathName);
}


BOOL ArasanGuiDoc::SaveModified() 
{
    // We don't generally edit game files .. so don't
    // prompt to change a saved document. The user can
    // still save manually, if desired (this is how
    // Arasan 5.4 and previous versions operated).
    
	return TRUE;
}
