// Copyright 1996-2000 by Jon Dart.  All Rights Reserved.
#ifndef _BITMAP_H
#define _BITMAP_H

#include "types.h"
#include "square.h"
#include <iostream>
#if defined (USE_INTRINSICS) && defined(_MSC_VER)
#include <intrin.h>
#endif

using namespace std;

#define FAST_SHIFT_R(data,n) \
    (n>=32) ? Bitmap(data.hivalue()>>(n)-32) \
         : Bitmap((uint32)data.value() >> (n))

#if defined(_MSC_VER) && _MSC_VER >= 1200
#define FORCEINLINE __forceinline
#else
#define FORCEINLINE __inline
#endif
         
#if defined(USE_ASM) && defined(__LP64__)
static FORCEINLINE unsigned int FirstOneAsm(const uint64 bits)
{
   uint64 ret;
   __asm__ (
            "bsfq %[bits], %[ret]"
            :[ret] "=r" (ret)
            :[bits] "mr" (bits)
	   );
   return (unsigned int)ret;
}
#endif

    struct masks
    {
       uint64 mask;
       uint64 maskDa1;
       uint64 maskDa8;
       uint64 maskR90;
       uint64 not_mask;
       uint64 not_maskDa1;
       uint64 not_maskDa8;
       uint64 not_maskR90;
    };
    
extern masks all_masks[64];

class Bitmap
{
       friend ostream & operator << (ostream &o, const Bitmap &b);

    public:	

       struct ints
       {
          uint32 loval,hival;
       };
       struct shorts
       {
          uint16 lo1,lo2,hi1,hi2;
       };
       union conv
       {
          uint64 val1;
          ints val2;
          shorts val3;
       };

    Bitmap()
    : data(0)
    {
    }
    
    Bitmap(uint64 n)
    : data(n)
    {
    }
    
    uint64 value() const
    {
       return data;
    }
    
    void *addr() const
    {
       return (void*)&data;
    }
    
    byte byte_value() const
    {
       // PC is little-endian
       return *((byte*)&data);
    }
    
    uint32 hivalue() const
    {
       return ((conv*)(this))->val2.hival;
    }
    
    uint32 lovalue() const
    {
       return ((conv*)(this))->val2.loval;
    }
    
    void clear()
    {
       data = 0;
    }
    void set_lsb()
    {
     data &= ~(data-1);
    }
    void set(int n)
    {
       data |= all_masks[n].mask;
    }
    void clear(int n)
    {
       data &= all_masks[n].not_mask;
    }
    void setR90(int n)
    {
        data |= all_masks[n].maskR90;
    }
    void clearR90(int n)
    {
        data &= all_masks[n].not_maskR90;
    }
    void setDa1(int n)
    {
        data |= all_masks[n].maskDa1;
    }
    void clearDa1(int n)
    {
        data &= all_masks[n].not_maskDa1;
    }
    void setDa8(int n)
    {
        data |= all_masks[n].maskDa8;
    }
    void clearDa8(int n)
    {
        data &= all_masks[n].not_maskDa8;
    }
    void setClear(const Bitmap &b)
    {
        data ^= b.data;
    }
    int is_set(int n) const
    {
        return ((data & mask[n]) != (uint64)0);
    }
    int is_clear()const
    {
        return (data == (uint64)0);
    }
    static Bitmap And( const Bitmap &src1, const Bitmap &src2)
    {
        return Bitmap(src1.data & src2.data);
    }
    void And( const Bitmap &b)
    {
        data &= b.data;
    }
    void andNot(const Bitmap &b)
    {
       data &= ~b.data;
    }
    static Bitmap andNot( const Bitmap &src1, const Bitmap &src2)
    {
       return Bitmap(src1.data & ~src2.data);
    }
    static Bitmap Or( const Bitmap &src1, const Bitmap &src2)
    {
       return Bitmap(src1.data | src2.data);
    }
    void Or( const Bitmap &b)
    {
        data |= b.data;
    }
    static inline Bitmap shr(const Bitmap &src, int n)
    {
        return Bitmap(src.data >> n);
        //uint32 *p = (uint32*)&src.data;
        //return Bitmap((uint32)(p[n/32] >> (n%32)));
    }
    static inline Bitmap fast_shr(const Bitmap &src, int n)
    {
        uint32 *p = (uint32*)&src.data;
        return Bitmap((uint32)(p[n/32] >> (n%32)));
    }
    void shr(int n)
    {
        data = data >> n;    
    }
    static Bitmap shl(const Bitmap &src, int n)
    {
        return Bitmap(src.data << n);
    }
    void shl(int n)
    {
        data = data << n;    
    }
    void shl_byte(int n)
    {
        // shift lowest 8 bits left by n bytes
        char *p = (char*)&data;
        p[n] = *p;
        if (n>0) *p = 0;
    }
    static inline Bitmap shl_byte(byte b, int n)
    {
        // shift lowest 8 bits left by n bytes
        uint64 new_data = 0;
        char *p = (char*)&new_data;
        p[n] = b;
        return Bitmap(new_data);
    }
    FORCEINLINE Bitmap shl_all(int n) const
    {
        // shift n bits, assuming that you will not
        // cross the 32-bit boundary
        return Bitmap(MAKELONGLONG(((conv*)&data)->val2.hival >> n,
          ((conv*)&data)->val2.loval  >> n));
    }
    
    int operator == (const Bitmap &b)
    {
       return b.data == data;
    }

    int operator != (const Bitmap &b)
    {
       return b.data != data;
    }

    // extract the bits for the nth rank (from the Black side)
    unsigned rank_bits(int n) const
    {
       byte *p = (byte*)&data;
       return (unsigned)p[n];
    }

    // return the number of bits set
    unsigned int bit_count() const
    {
#if defined(USE_ASM) && defined(_MSC_VER)
	   return BitCountAsm(data);
#else
       unsigned register int n;
       uint64 tmp = data;
       for (n = 0; tmp != 0; n++, tmp &= (tmp - 1));
       return n;
#endif
    }
    
    Square first_one() const
    {
      if (data == 0) return InvalidSquare;
#if defined(USE_ASM) && defined(__LP64__)
      return FirstOneAsm(data);
#elif defined(_WIN64) && defined(_MSC_VER) && defined(USE_INTRINSICS)
      DWORD index;
      if (_BitScanForward64(&index,data))
        return (Square)index;
      else 
        return InvalidSquare;
#else
      extern int *first_ones;
      register uint32 copy;
      int word = (lovalue() == 0);
      copy = ((uint32*)&data)[word];
      copy &= ~(copy-1);
      if (copy & 0xffff)
           return first_ones[copy & 0xffff] + word*32;
      else 
           return first_ones[copy >> 16]+word*32+16;
#endif
    }
    
    static void init();
    
    static void cleanup();

    static const uint64 mask[64];
    static int unrot90[64];
    static int R90[64];
    static int da1[64];
    static int da1mod8[64];
    static int da1div8[64];
    static int unrot_da1[64];
    static int da1_mask[64];
    static int da8[64];
    static int da8mod8[64];
    static int da8div8[64];
    static int unrot_da8[64];
    static int da8_mask[64];
    
    uint64 data;

};

#define DIAG_UPPER_LEFT 0
#define DIAG_UPPER_RIGHT 1
#define DIAG_LOWER_LEFT 2
#define DIAG_LOWER_RIGHT 3

struct DiagInfo2
{
   int shift;
   int byte_shift;
   int bit_shift;
   int bit;
   int mask;
   Bitmap *attacks;
   Bitmap upperMask;
   Bitmap lowerMask;
};

extern DiagInfo2 DiagInfoA1[64], DiagInfoA8[64];

inline int TEST_MASK(const Bitmap &b1,const Bitmap &b2) {
  return !Bitmap::And(b1,b2).is_clear();
}

inline uint64 And( const Bitmap &src1, const Bitmap &src2)
{
    return (src1.data & src2.data);
}

inline uint64 Or( const Bitmap &src1, const Bitmap &src2)
{
    return (src1.data | src2.data);
}
#endif
