#include "draw.h"
#include "util.h"
#include "bitmap.h"
#include "bearing.h"

int repetition_draw( const Board &board )
{
   return board.rep_count() >= 2;
}

int material_draw( const Board &board)
{
   // check for insufficient material
   const Material &mat1 = board.getMaterial(White);
   const Material &mat2 = board.getMaterial(Black);
   if (mat1.pawn_count() || mat2.pawn_count())
   {
      return 0 /*FALSE*/;
   }
   if ((mat1.value() <= KING_VALUE + BISHOP_VALUE) &&
       (mat2.value() <= KING_VALUE + BISHOP_VALUE))
   {
      if (mat1.king_only() || mat2.king_only())
         // K vs K, or K vs KN, or K vs KB
         return 1 /*TRUE*/;
      else if (mat1.knight_count() || mat2.knight_count())
      {
         // KN vs KN 
         return 0 /*FALSE*/;
      } else
      {
         Bitmap wbishops(board.bishop_bits[White]);
         Bitmap bbishops(board.bishop_bits[Black]);
         // drawn if same-color bishops
         if
             (!(Bitmap::and(wbishops,Bearing::black_squares)).is_clear())
         {
            return !(Bitmap::and(bbishops,Bearing::black_squares)).is_clear();
         } else
         {
            return !(Bitmap::and(bbishops,Bearing::white_squares)).is_clear();
         }
      }
   }
   return 0 /*FALSE*/;
}

int is_draw(const Board &board)
{
   return (repetition_draw(board) || material_draw(board) ||
           is_theoretical_draw(board));
}        

int is_legal_draw(const Board &board)
{
   return (repetition_draw(board) || material_draw(board));
}        

int is_theoretical_draw(const Board &board)
{
   const Material &mat1 = board.getMaterial(White);
   const Material &mat2 = board.getMaterial(Black);
   if (mat1.value() > 2000+(BISHOP_VALUE*2) &&
       mat2.value() > 2000+(BISHOP_VALUE*2))
       return 0 /*FALSE*/;
   // Check for K + P vs rook pawn
   if ((unsigned)mat1.infobits() == Material::KP &&
       mat2.king_only())
   {
      if (!Bitmap::and(rook_pawn_mask,
                       board.pawn_bits[White]).is_clear())
      {
         Square kp = board.KingPos(White);
         Square kp2 = board.KingPos(Black);
         Square psq = (board.pawn_bits[White].first_one());
         if (Rank(kp,White) < Rank(kp2,White))
         {
            Square qsq = Square(File(psq),8,White);
            return !uncatchable(board,psq,kp,kp2,White) &&
            distance1(kp2,qsq) < distance1(kp,qsq);
         }
      }
   } 
   else if ((unsigned)mat2.infobits() == Material::KP &&
              mat1.king_only())
   {
      if (!Bitmap::and(rook_pawn_mask,
                       board.pawn_bits[Black]).is_clear())
      {
         Square kp = board.KingPos(Black);
         Square kp2 = board.KingPos(White);
         Square psq(board.pawn_bits[Black].first_one());
         if (Rank(kp,Black) < Rank(kp2,Black))
         {
            Square qsq = Square(File(psq),8,Black);
            return !uncatchable(board,psq,kp,kp2,Black) &&
            distance1(kp2,qsq) < distance1(kp,qsq);
         }
      }
   }
   // Check for wrong bishop + rook pawn vs king.  Not very common but
   // we don't want to get it wrong.
   else if (mat1.infobits() == Material::KBP && mat2.king_only())
   {
      Bitmap pb(board.pawn_bits[White]);
      if (!(Bitmap::and(pb,Bearing::file_mask[0])).is_clear())
      {
         // on the a file
         // queen square is white
         if ((Bitmap::and(board.bishop_bits[White],Bearing::white_squares)).is_clear())
         {
            // wrong color bishop
            Square pawnsq = pb.first_one();
            if (Rank(board.KingPos(Black),White) >= Rank(pawnsq,White))
            {
               if (Util::Abs((int)File(board.KingPos(Black)) - (int)File(pawnsq)) <= 1)
               {
                  return 1 /*TRUE*/;
               }
            }
         }
      } 
      else if (!(Bitmap::and(pb,Bearing::file_mask[7])).is_clear())
      {
         // on the h file
         // queen square is black
         if ((Bitmap::and(board.bishop_bits[White],Bearing::black_squares)).is_clear())
         {
             // wrong color bishop
             Square pawnsq = pb.first_one();
             if (Rank(board.KingPos(Black),White) >= Rank(pawnsq,White))
             {
                if (Util::Abs((int)File(board.KingPos(Black)) - (int)File(pawnsq)) <= 1)
                {
                   return 1 /*TRUE*/;
                }
             }
         }
      }
   } 
   else if (mat2.infobits() == Material::KBP && mat1.king_only())
   {
      Bitmap pb(board.pawn_bits[Black]);
      if (!(Bitmap::and(pb,Bearing::file_mask[0])).is_clear())
      {
         // on the a file
         // queen square is black
         if ((Bitmap::and(board.bishop_bits[Black],Bearing::black_squares)).is_clear())
         {
            // wrong color bishop
            Square pawnsq = pb.first_one();
            if (Rank(board.KingPos(White),Black) >= Rank(pawnsq,Black))
            {
               if (Util::Abs((int)File(board.KingPos(White)) - (int)File(pawnsq)) <= 1)
               {
                  return 1 /*TRUE*/;
               }
            }
         }
      } 
      else if (!(Bitmap::and(pb,Bearing::file_mask[7])).is_clear())
      {
         // on the h file
         // queen square is white
         if ((Bitmap::and(board.bishop_bits[Black],Bearing::white_squares)).is_clear())
         {
            // wrong color bishop
            Square pawnsq = pb.first_one();
            if (Rank(board.KingPos(White),Black) >= Rank(pawnsq,Black))
            {
               if (Util::Abs((int)File(board.KingPos(White)) - (int)File(pawnsq)) <= 1)
               {
                  return 1 /*TRUE*/;
               }
            }
         }
      }
   }
   // check for KNN vs K
   if (((unsigned)mat1.infobits() == Material::KNN &&
        mat2.king_only()) ||
       ((unsigned)mat2.infobits() == Material::KNN &&
        mat1.king_only()))
      return 1 /*TRUE*/;

   return 0 /*FALSE*/;
}

