// Copyright 1994-2002 by Jon Dart.  All Rights Reserved.

// Stand-alone console executable to test the search engine

#include "board.h"
#include "movegen.h"
#include "search.h"
#include "movearr.h"
#include "notation.h"
#include "globals.h"
#include "chessio.h"
#include "epdrec.h"
#include "scoring.h"
#include "bearing.h"
#include "calctime.h"
#include <string>

extern "C"
{
#include <string.h>
#include <ctype.h>
};
#include <fstream>
#include <iostream>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#ifdef TABLEBASES
#include "tbprobe.h"
#endif

int verbose = 0;
long total_nodes = 0L;
long total_correct = 0L;
long total_time = 0L;
int early_exit_plies = Constants::MaxPly;
int early_exit = 0;
int solution_move_count = 0;
int solution_time = 0;
int last_iteration_depth = -1;
int iterations_correct = 0;
Move solution_moves[10];
bool avoid = false;

static void post(const Statistics &stats)
{
   if (avoid) return;
   for (int i = 0; i < solution_move_count; ++i)
   {
      if (MovesEqual(solution_moves[i],stats.best_line[0]))
      {
         if ((int)stats.depth > last_iteration_depth)
         {
            ++iterations_correct;
            last_iteration_depth = stats.depth;
         }
         // don't terminate in under 2 seconds unless we found
         // a mate
         if ((stats.elapsed_time >= 200 ||
              stats.value > Constants::BIG-100)
             && iterations_correct >= early_exit_plies)
	   early_exit = 1;
         if (solution_time == -1) solution_time = stats.elapsed_time;
         return;
      }
   }
   iterations_correct = 0;
}

static int terminate(const Statistics &stats)
{
   post(stats);
   return early_exit;
}

static Move search(Board &board, int ply_limit,
                   int time_limit, Statistics &stats)
{
   Search searcher;

   searcher.register_post_function(post);
   searcher.register_terminate_function(terminate);
   solution_time = -1;

   Move move = searcher.find_best_move(board, 
                                       ply_limit >= 99 ? Search::Fixed_Time : Search::Fixed_Depth, 
                                       time_limit, 0, ply_limit,
                                       0, stats,
                                       verbose ? Search::Debug : Search::Silent);

   char move_text[20];
   Notation::Image(board, move, move_text);
   cout << "search:";
   cout << '\t' << move_text << "\t" << stats.elapsed_time/100 <<
   " seconds.\tscore:" << Scoring::print_score(stats.display_value) << "\t"  <<
   stats.num_nodes << " nodes.";
   total_time += stats.elapsed_time;
   total_nodes += stats.num_nodes;

   game_moves->removeAll();
//    MoveRecord::freeAll(TRUE);

   return move;
}

static void show_usage()
{
   cerr << "Usage:"  << endl;
   cerr << "  testsrc <options> <position file>" << endl;
   cerr << "Options are:" << endl;
   cerr << "  -e           - show root evaluation only" << endl;    
   cerr << "  -P           - performance test" << endl;    
   cerr << "  -t [seconds] - seconds to search per position" << endl;
   cerr << "  -p [plies]   - ply limit for search" << endl;
   cerr << "  -n [0|1|2]   - null depth reduction (default 2)" << endl;
   cerr << "  -k [0|1|2]   - killer moves per ply (default 2)" <<
   endl;
   cerr << "  -H [size]    - hash table size (kilobytes)" << endl;
   cerr << "  -c[+|-]      - use/don't use check extensions (default on)" << endl;
   cerr << "  -r[+|-]      - use/don't use recapture extensions (default on)" << endl;
   cerr << "  -pp[+|-]     - use/don't use pawn push extensions (default on)" << endl;
   cerr  << "  -tb[+|-]     - use/don't use tablebases (default on)" << endl;
   cerr << "  -h[+|-]      - use/don't use history heuristic (default on)" << endl;
   cerr << "  -x [plies]   - terminate if correct move holds for <plies>" << endl;
   cerr << "  -v           - be verbose\n" << endl;
}

static void do_eval(ifstream &pos_file, Board &board)
{
   pos_file >> board;
   if (!pos_file.good())
   {
      cout << "File not found, or bad format." << endl;
      return;
   }
#ifdef TABLEBASES
   int tbscore;
   if (options.search.use_tablebases && probe_tb(board,tbscore,0)) {
     static char line[128];
     sprintf(line,"score = %s (from tablebases)\n",Scoring::print_score(tbscore));
     cout << line << endl;
   }
   else 
#endif
   {
     Scoring::init(board);
     static char line[128];
     sprintf(line,"score = %s\n",Scoring::print_score(Scoring::evalu8(board)));
     cout << line << endl;
   }
}

static void do_perf(ifstream &pos_file, Board &board)
{
   pos_file >> board;
   if (!pos_file.good())
   {
      cout << "File not found, or bad format." << endl;
      return;
   }
   Scoring::init(board);
   Board_State save_state(board);
   Move_Generator mg(board,0,NullMove);
   Move move_list[Move_Generator::MaxMoves];
   int current_time = getTime();
   long total = 0;
   int i;
   for (i = 0; i < 2000000; i++)
   {
      total += mg.Generate_Moves(move_list);
   }
   int new_time = getTime();
   float mg_time = ((float)(new_time-current_time))/100.0F;
   cout << "movegen:  " <<  ((float)total)/mg_time << " generations/sec."
   << endl;
   current_time = new_time;
   for (i = 0; i < 200000; i++)
   {
      int n = mg.Generate_Moves(move_list);
      for (int j=0;j<n;j++)
      {
        board.DoMove(move_list[j]);
        board.UndoMove(move_list[j],save_state);
      }
   }
   new_time = getTime();
   float mg_make_time = ((float)(new_time-current_time))/100.0F;
   cout << "movegen/make/unmake: " << ((float)total/10.0F)/mg_make_time << " per sec." <<endl;
   cout << endl;
}

static const char *get_move(const char *buf,const Board &board, Move &m)
{
   const char *p = buf;
   m = NullMove;
   if (*p)
   {
      while (isspace(*p) && *p != '\0') ++p;
      if (*p == '\0')
         return NULL;
      char tmp[10];
      int i = 0;
      char *q = tmp;
      while (!isspace(*p) && *p != '+' && *p != '\0' && i < 10)
      {
         *q++ = *p++;
         ++i;
      }
      *q = '\0'; 
      m = Notation::Value(board,board.Side(),tmp);
      if (*p == '+') ++p;
   }
   return p;
}

static void do_search(Board &board,int ply_limit,int time_limit)
{
   Statistics stats;
   Move m = search(board,ply_limit,time_limit,stats);

   int correct = avoid ? 1 : 0;
   for (int i = 0; i < solution_move_count; ++i)
   {
      if (MovesEqual(solution_moves[i],m))
      {
         correct = (avoid) ? 0 : 1;
         break;
      }
   }
   if (correct)
   {
      total_correct++;
      cout << "\t++ correct" << endl;
   }
   else
      cout << "\t** error" << endl;
   cout << stats.best_line_image << endl;
}

static void do_search(ifstream &pos_file, Board &board, int ply_limit,
                      int time_limit) 
{
   char buf[512];
   while (!pos_file.eof())
   {
      pos_file.getline(buf,511);
      if (!pos_file)
      {
         cout << "File not found, or bad format." << endl;
         return;
      }
      // Try to parse this line as an EPD command.
#if defined(__GNUC__) && _GNUC_PREREQ(3,2)
      string istr(buf);
      istringstream stream(istr);
#else
      istrstream stream(buf,512);
#endif
      EPDRecord *epd_rec =
        ChessIO::readEPDRecord(stream,board);
      if (epd_rec != NULL)
      {
         for (int i = 0; i < epd_rec->getSize(); i++)
         {
            string key, val;
            epd_rec->getData(i,key,val);
            if (key == "bm" || key == "am")
            {
               Move m;
               solution_move_count = 0;
               const char *p = val.c_str();
               while (*p) {
                  p = get_move(p,board,m);
                  if (!IsNull(m))
                  {
                     assert(solution_move_count < 10);
                     solution_moves[solution_move_count++] = m;
                  }
                  avoid = (key == "am");
               }
	    }
            else if (key == "id")
            {
               cout << val << endl;
            }
         }
         last_iteration_depth = -1;
         iterations_correct = 0;
         early_exit = 0;
         do_search(board,ply_limit,time_limit);
      }

      char c;
      while (!pos_file.eof())
      {
         c = pos_file.get();
         if (!isspace(c) && c != '\n')
         {
            if (!pos_file.eof())
               pos_file.putback(c);
            break;
         }
      }
   }
   pos_file.close();
   cout << endl; 
   cout << "total time     : " << total_time/100 << " seconds."
   <<endl;
   cout << "total nodes    : " << total_nodes << endl;
   cout << "number correct : " << total_correct << endl;
}

int CDECL main(int argc, char **argv)
{
   init_globals(argv[0],0,0,1);
   atexit(cleanup_globals);

   Board board;
   int ply_limit = 2;
   int time_limit = 99999;
   int arg = 1;
   int show_eval = 0;
   int perf = 0;

   if (argc == 1)
   {
      show_usage();
      return -1;
   }
   else
   {
      while (arg < argc && *(argv[arg]) == '-')
      {
         char c = *(argv[arg]+1);
         switch (c)
         {
         case 'e':
            ++show_eval;
            break;
         case 'P':
            ++perf;
            break;
         case 'p':
            c = tolower(*(argv[arg]+2));
            if (c == 'p')
            {
               c = tolower(*(argv[arg]+3));
               int pawn_push_extensions = (c=='\0' || c=='+');
               options.search.pawn_push_extensions = pawn_push_extensions;
            }
            else
            {
               ++arg;
               ply_limit = atoi(argv[arg]);
            }
            break;
         case 't': {
	        c = (*(argv[arg]+3));
            if (c == 'b') {
               c = (*(argv[arg]+4));
               if (c == '\0' || c == '+') {
 		          options.search.use_tablebases = 1;
               } else {
		          options.search.use_tablebases = 0;
               }
            }
            else {
               ++arg;
               time_limit = 100*atol(argv[arg]);
               ply_limit = 99;
            }
            }
            break;
         case 'v':
            verbose++; break;
         case 'n':
            ++arg;
            if (arg < argc)
            {
               if (argv[arg][0] == '0')
                  options.search.null_depth_reduction = 0;
               else if (argv[arg][0] == '1')
                  options.search.null_depth_reduction = 1;
               else if (argv[arg][0] == '2')
                  options.search.null_depth_reduction = 2;
            }
            break;
         case 'H':
            {
               ++arg;
               options.search.hash_table_size = atol(argv[arg]);
            }
            break;
         case 'k':
            ++arg;
            if (arg < argc)
            {
               int killers = -1;
               if (argv[arg][0] == '0')
                  killers = 0;
               else if (argv[arg][0] == '1')
                  killers = 1;
               else if (argv[arg][0] == '2')
                  killers = 2;
               if (killers >= 0)
               {
                  options.search.num_killers = killers;
               }
            }
            break;
         case 'c':
            c = tolower(*(argv[arg]+2));
            {
               int check_extensions = (c=='\0' || c=='+');
               options.search.check_extensions = check_extensions;
            }
            break;
         case 'f':
            c = tolower(*(argv[arg]+2));
            {
               int forced_extensions = (c=='\0' || c=='+');
               options.search.forced_extensions = forced_extensions;
            }
            break;
         case 'r':
            c = tolower(*(argv[arg]+2));
            {
               int recapture_extensions = (c=='\0' || c=='+');
               options.search.recapture_extensions = recapture_extensions;
            }
            break;
         case 'x':
            {
               ++arg;
               if (arg < argc)
               {
                  early_exit_plies = atoi(argv[arg]);
               }
               break;
            }
         case 'h':
            c = tolower(*(argv[arg]+2));
            {
               int history = (c=='\0' || c=='+');
               options.search.use_history = history;
            }
            break;
         default:
            show_usage();
            return -1;
         }
         ++arg;
      }

      if (arg >= argc)
      {
         show_usage();
         return -1;
      }
   }

#if defined (_WIN32) || (defined (__GNUC__) && (__GNUC__ >=3))
      ifstream pos_file( argv[arg], ios::in);
#else
      ifstream pos_file( argv[arg], ios::in | ios::nocreate);
#endif

   if (pos_file.good())
   {
      if (show_eval)
         do_eval(pos_file,board);
      else if (perf)
         do_perf(pos_file,board);
      else
         do_search(pos_file,board,ply_limit,time_limit);         
   }
   else
   {
      cout << "file not found: " << argv[arg] << endl;
      return -1;
   }
   return 0;
}
